#    Copyright 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import threading

from concurrent import futures
import futurist

import oslo_messaging
from oslo_messaging._drivers import common as rpc_common
from oslo_messaging._drivers.zmq_driver.client.publishers\
    import zmq_publisher_base
from oslo_messaging._drivers.zmq_driver import zmq_address
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names
from oslo_messaging._drivers.zmq_driver import zmq_socket

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


class DealerCallPublisher(zmq_publisher_base.PublisherBase):
    """Thread-safe CALL publisher

        Used as faster and thread-safe publisher for CALL
        instead of ReqPublisher.
    """

    def __init__(self, conf, matchmaker):
        super(DealerCallPublisher, self).__init__(conf)
        self.matchmaker = matchmaker
        self.reply_waiter = ReplyWaiter(conf)
        self.sender = RequestSender(conf, matchmaker, self.reply_waiter) \
            if not conf.zmq_use_broker else \
            RequestSenderLight(conf, matchmaker, self.reply_waiter)

    def send_request(self, request):
        reply_future = self.sender.send_request(request)
        try:
            reply = reply_future.result(timeout=request.timeout)
        except futures.TimeoutError:
            raise oslo_messaging.MessagingTimeout(
                "Timeout %s seconds was reached" % request.timeout)
        finally:
            self.reply_waiter.untrack_id(request.message_id)

        LOG.debug("Received reply %s" % reply)
        if reply[zmq_names.FIELD_FAILURE]:
            raise rpc_common.deserialize_remote_exception(
                reply[zmq_names.FIELD_FAILURE],
                request.allowed_remote_exmods)
        else:
            return reply[zmq_names.FIELD_REPLY]


class RequestSender(zmq_publisher_base.PublisherMultisend):

    def __init__(self, conf, matchmaker, reply_waiter):
        super(RequestSender, self).__init__(conf, matchmaker, zmq.DEALER)
        self.reply_waiter = reply_waiter
        self.queue, self.empty_except = zmq_async.get_queue()
        self.executor = zmq_async.get_executor(self.run_loop)
        self.executor.execute()

    def send_request(self, request):
        reply_future = futurist.Future()
        self.reply_waiter.track_reply(reply_future, request.message_id)
        self.queue.put(request)
        return reply_future

    def _do_send_request(self, socket, request):
        socket.send(b'', zmq.SNDMORE)
        socket.send_pyobj(request)

        LOG.debug("Sending message_id %(message)s to a target %(target)s"
                  % {"message": request.message_id,
                     "target": request.target})

    def _check_hosts_connections(self, target, listener_type):
        if str(target) in self.outbound_sockets:
            socket = self.outbound_sockets[str(target)]
        else:
            hosts = self.matchmaker.get_hosts(
                target, listener_type)
            socket = zmq_socket.ZmqSocket(self.zmq_context, self.socket_type)
            self.outbound_sockets[str(target)] = socket

            for host in hosts:
                self._connect_to_host(socket, host, target)

        return socket

    def run_loop(self):
        try:
            request = self.queue.get(timeout=self.conf.rpc_poll_timeout)
        except self.empty_except:
            return

        socket = self._check_hosts_connections(
            request.target, zmq_names.socket_type_str(zmq.ROUTER))

        self._do_send_request(socket, request)
        self.reply_waiter.poll_socket(socket)


class RequestSenderLight(RequestSender):
    """This class used with proxy.

        Simplified address matching because there is only
        one proxy IPC address.
    """

    def __init__(self, conf, matchmaker, reply_waiter):
        if not conf.zmq_use_broker:
            raise rpc_common.RPCException("RequestSenderLight needs a proxy!")

        super(RequestSenderLight, self).__init__(
            conf, matchmaker, reply_waiter)

        self.socket = None

    def _check_hosts_connections(self, target, listener_type):
        if self.socket is None:
            self.socket = zmq_socket.ZmqSocket(self.zmq_context,
                                               self.socket_type)
            self.outbound_sockets[str(target)] = self.socket
            address = zmq_address.get_broker_address(self.conf)
            self._connect_to_address(self.socket, address, target)
        return self.socket

    def _do_send_request(self, socket, request):
        LOG.debug("Sending %(type)s message_id %(message)s"
                  " to a target %(target)s"
                  % {"type": request.msg_type,
                     "message": request.message_id,
                     "target": request.target})

        envelope = request.create_envelope()

        socket.send(b'', zmq.SNDMORE)
        socket.send_pyobj(envelope, zmq.SNDMORE)
        socket.send_pyobj(request)


class ReplyWaiter(object):

    def __init__(self, conf):
        self.conf = conf
        self.replies = {}
        self.poller = zmq_async.get_poller()
        self.executor = zmq_async.get_executor(self.run_loop)
        self.executor.execute()
        self._lock = threading.Lock()

    def track_reply(self, reply_future, message_id):
        self._lock.acquire()
        self.replies[message_id] = reply_future
        self._lock.release()

    def untrack_id(self, message_id):
        self._lock.acquire()
        self.replies.pop(message_id)
        self._lock.release()

    def poll_socket(self, socket):

        def _receive_method(socket):
            empty = socket.recv()
            assert empty == b'', "Empty expected!"
            reply = socket.recv_pyobj()
            LOG.debug("Received reply %s" % reply)
            return reply

        self.poller.register(socket, recv_method=_receive_method)

    def run_loop(self):
        reply, socket = self.poller.poll(
            timeout=self.conf.rpc_poll_timeout)
        if reply is not None:
            call_future = self.replies[reply[zmq_names.FIELD_MSG_ID]]
            call_future.set_result(reply)




# Copyright 2013 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

import six

from oslo_config import cfg
from oslo_messaging import exceptions

base_opts = [
    cfg.IntOpt('rpc_conn_pool_size',
               default=30,
               deprecated_group='DEFAULT',
               help='Size of RPC connection pool.'),
]


class TransportDriverError(exceptions.MessagingException):
    """Base class for transport driver specific exceptions."""


@six.add_metaclass(abc.ABCMeta)
class IncomingMessage(object):

    def __init__(self, listener, ctxt, message):
        self.conf = listener.conf
        self.listener = listener
        self.ctxt = ctxt
        self.message = message

    @abc.abstractmethod
    def reply(self, reply=None, failure=None, log_failure=True):
        "Send a reply or failure back to the client."

    def acknowledge(self):
        "Acknowledge the message."

    @abc.abstractmethod
    def requeue(self):
        "Requeue the message."


@six.add_metaclass(abc.ABCMeta)
class Listener(object):

    def __init__(self, driver):
        self.conf = driver.conf
        self.driver = driver

    @abc.abstractmethod
    def poll(self, timeout=None):
        """Blocking until a message is pending and return IncomingMessage.
        Return None after timeout seconds if timeout is set and no message is
        ending or if the listener have been stopped.
        """

    def stop(self):
        """Stop listener.
        Stop the listener message polling
        """
        pass

    def cleanup(self):
        """Cleanup listener.
        Close connection (socket) used by listener if any.
        As this is listener specific method, overwrite it in to derived class
        if cleanup of listener required.
        """
        pass


@six.add_metaclass(abc.ABCMeta)
class BaseDriver(object):

    def __init__(self, conf, url,
                 default_exchange=None, allowed_remote_exmods=None):
        self.conf = conf
        self._url = url
        self._default_exchange = default_exchange
        self._allowed_remote_exmods = allowed_remote_exmods or []

    def require_features(self, requeue=False):
        if requeue:
            raise NotImplementedError('Message requeueing not supported by '
                                      'this transport driver')

    @abc.abstractmethod
    def send(self, target, ctxt, message,
             wait_for_reply=None, timeout=None, envelope=False):
        """Send a message to the given target."""

    @abc.abstractmethod
    def send_notification(self, target, ctxt, message, version):
        """Send a notification message to the given target."""

    @abc.abstractmethod
    def listen(self, target):
        """Construct a Listener for the given target."""

    @abc.abstractmethod
    def listen_for_notifications(self, targets_and_priorities, pool):
        """Construct a notification Listener for the given list of
        tuple of (target, priority).
        """

    @abc.abstractmethod
    def cleanup(self):
        """Release all resources."""



#    Copyright 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from oslo_messaging._drivers.zmq_driver.client.publishers.dealer \
    import zmq_dealer_publisher
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names
from oslo_messaging._i18n import _LI, _LW

zmq = zmq_async.import_zmq()

LOG = logging.getLogger(__name__)


class DealerPublisherProxy(zmq_dealer_publisher.DealerPublisher):

    def __init__(self, conf, matchmaker, reply_receiver):
        super(DealerPublisherProxy, self).__init__(conf, matchmaker)
        self.reply_receiver = reply_receiver

    def send_request(self, multipart_message):

        envelope = multipart_message[zmq_names.MULTIPART_IDX_ENVELOPE]

        LOG.debug("Envelope: %s" % envelope)

        target = envelope[zmq_names.FIELD_TARGET]
        dealer_socket = self._check_hosts_connections(
            target, zmq_names.socket_type_str(zmq.ROUTER))

        if not dealer_socket.connections:
            # NOTE(ozamiatin): Here we can provide
            # a queue for keeping messages to send them later
            # when some listener appears. However such approach
            # being more reliable will consume additional memory.
            LOG.warning(_LW("Request %s was dropped because no connection")
                        % envelope[zmq_names.FIELD_MSG_TYPE])
            return

        self.reply_receiver.track_socket(dealer_socket.handle)

        LOG.debug("Sending message %(message)s to a target %(target)s"
                  % {"message": envelope[zmq_names.FIELD_MSG_ID],
                     "target": envelope[zmq_names.FIELD_TARGET]})

        if envelope[zmq_names.FIELD_MSG_TYPE] in zmq_names.MULTISEND_TYPES:
            for _ in range(dealer_socket.connections_count()):
                self._send_request(dealer_socket, multipart_message)
        else:
            self._send_request(dealer_socket, multipart_message)

    def _send_request(self, socket, multipart_message):

        socket.send(b'', zmq.SNDMORE)
        socket.send_pyobj(
            multipart_message[zmq_names.MULTIPART_IDX_ENVELOPE],
            zmq.SNDMORE)
        socket.send(multipart_message[zmq_names.MULTIPART_IDX_BODY])


class ReplyReceiver(object):

    def __init__(self, poller):
        self.poller = poller
        LOG.info(_LI("Reply waiter created in broker"))

    def _receive_reply(self, socket):
        return socket.recv_multipart()

    def track_socket(self, socket):
        self.poller.register(socket, self._receive_reply)

    def cleanup(self):
        self.poller.close()



#    Copyright 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import logging

import six

from oslo_messaging._drivers import common as rpc_common
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names
from oslo_messaging._drivers.zmq_driver import zmq_socket
from oslo_messaging._i18n import _LE

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


@six.add_metaclass(abc.ABCMeta)
class ConsumerBase(object):

    def __init__(self, conf, poller, server):
        self.conf = conf
        self.poller = poller
        self.server = server
        self.sockets = []
        self.context = zmq.Context()

    def subscribe_socket(self, socket_type):
        try:
            socket = zmq_socket.ZmqRandomPortSocket(
                self.conf, self.context, socket_type)
            self.sockets.append(socket)
            self.poller.register(socket, self.receive_message)
            LOG.debug("Run %(stype)s consumer on %(addr)s:%(port)d",
                      {"stype": zmq_names.socket_type_str(socket_type),
                       "addr": socket.bind_address,
                       "port": socket.port})
            return socket
        except zmq.ZMQError as e:
            errmsg = _LE("Failed binding to port %(port)d: %(e)s")\
                % (self.port, e)
            LOG.error(_LE("Failed binding to port %(port)d: %(e)s")
                      % (self.port, e))
            raise rpc_common.RPCException(errmsg)

    @abc.abstractmethod
    def listen(self, target):
        """Associate new sockets with targets here"""

    @abc.abstractmethod
    def receive_message(self, target):
        """Method for poller - receiving message routine"""

    def cleanup(self):
        for socket in self.sockets:
            if not socket.handle.closed:
                socket.setsockopt(zmq.LINGER, 0)
                socket.close()
        self.sockets = []


class SingleSocketConsumer(ConsumerBase):

    def __init__(self, conf, poller, server, socket_type):
        super(SingleSocketConsumer, self).__init__(conf, poller, server)
        self.socket = self.subscribe_socket(socket_type)

    @property
    def address(self):
        return self.socket.bind_address

    @property
    def port(self):
        return self.socket.port



# Copyright 2014 eNovance.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import aioeventlet
import trollius

from oslo_messaging._executors import impl_eventlet


class AsyncioEventletExecutor(impl_eventlet.EventletExecutor):

    """A message executor which integrates with eventlet and trollius.

    The executor is based on eventlet executor and so is compatible with it.
    The executor supports trollius coroutines, explicit asynchronous
    programming, in addition to eventlet greenthreads, implicit asynchronous
    programming.

    To use the executor, an aioeventlet event loop must the running in the
    thread executing the executor (usually the main thread). Example of code to
    setup and run an aioeventlet event loop for the executor (in the main
    thread)::

        import aioeventlet
        import trollius

        policy = aioeventlet.EventLoopPolicy()
        trollius.set_event_loop_policy(policy)

        def run_loop(loop):
            loop.run_forever()
            loop.close()

        # Get the aioeventlet event loop (create it if needed)
        loop = trollius.get_event_loop()

        # run the event loop in a new greenthread,
        # close it when it is done
        eventlet.spawn(run_loop, loop)

    """

    def __init__(self, conf, listener, dispatcher):
        super(AsyncioEventletExecutor, self).__init__(conf, listener,
                                                      dispatcher)
        self._loop = None

    def start(self):
        # check that the event loop is an aioeventlet event loop
        loop = trollius.get_event_loop()
        if not isinstance(loop, aioeventlet.EventLoop):
            raise RuntimeError("need an aioeventlet event loop")
        self._loop = loop

        super(AsyncioEventletExecutor, self).start()

    def _coroutine_wrapper(self, func, *args, **kw):
        result = func(*args, **kw)
        if trollius.iscoroutine(result):
            result = aioeventlet.yield_future(result, loop=self._loop)
        return result

    _executor_callback = _coroutine_wrapper




# Copyright 2013 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import threading

LOG = logging.getLogger(__name__)


def version_is_compatible(imp_version, version):
    """Determine whether versions are compatible.

    :param imp_version: The version implemented
    :param version: The version requested by an incoming message.
    """
    version_parts = version.split('.')
    imp_version_parts = imp_version.split('.')
    try:
        rev = version_parts[2]
    except IndexError:
        rev = 0
    try:
        imp_rev = imp_version_parts[2]
    except IndexError:
        imp_rev = 0

    if int(version_parts[0]) != int(imp_version_parts[0]):  # Major
        return False
    if int(version_parts[1]) > int(imp_version_parts[1]):  # Minor
        return False
    if (int(version_parts[1]) == int(imp_version_parts[1]) and
            int(rev) > int(imp_rev)):  # Revision
        return False
    return True


class DispatcherExecutorContext(object):
    """Dispatcher executor context helper

    A dispatcher can have work to do before and after the dispatch of the
    request in the main server thread while the dispatcher itself can be
    done in its own thread.

    The executor can use the helper like this:

        callback = dispatcher(incoming)
        callback.prepare()
        thread = MyWhateverThread()
        thread.on_done(callback.done)
        thread.run(callback.run)

    """
    def __init__(self, incoming, dispatch, executor_callback=None,
                 post=None):
        self._result = None
        self._incoming = incoming
        self._dispatch = dispatch
        self._post = post
        self._executor_callback = executor_callback

    def run(self):
        """The incoming message dispath itself

        Can be run in an other thread/greenlet/corotine if the executor is
        able to do it.
        """
        try:
            self._result = self._dispatch(self._incoming,
                                          self._executor_callback)
        except Exception:
            msg = 'The dispatcher method must catches all exceptions'
            LOG.exception(msg)
            raise RuntimeError(msg)

    def done(self):
        """Callback after the incoming message have been dispathed

        Should be ran in the main executor thread/greenlet/corotine
        """
        # FIXME(sileht): this is not currently true, this works only because
        # the driver connection used for polling write on the wire only to
        # ack/requeue message, but what if one day, the driver do something
        # else
        if self._post is not None:
            self._post(self._incoming, self._result)


def fetch_current_thread_functor():
    # Until https://github.com/eventlet/eventlet/issues/172 is resolved
    # or addressed we have to use complicated workaround to get a object
    # that will not be recycled; the usage of threading.current_thread()
    # doesn't appear to currently be monkey patched and therefore isn't
    # reliable to use (and breaks badly when used as all threads share
    # the same current_thread() object)...
    try:
        import eventlet
        from eventlet import patcher
        green_threaded = patcher.is_monkey_patched('thread')
    except ImportError:
        green_threaded = False
    if green_threaded:
        return lambda: eventlet.getcurrent()
    else:
        return lambda: threading.current_thread()


class DummyLock(object):
    def acquire(self):
        pass

    def release(self):
        pass

    def __enter__(self):
        self.acquire()

    def __exit__(self, type, value, traceback):
        self.release()



#    Copyright 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from oslo_messaging._drivers.zmq_driver.client.publishers\
    import zmq_publisher_base
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names
from oslo_messaging._i18n import _LW

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


class PushPublisher(zmq_publisher_base.PublisherMultisend):

    def __init__(self, conf, matchmaker):
        super(PushPublisher, self).__init__(conf, matchmaker, zmq.PUSH)

    def send_request(self, request):

        if request.msg_type == zmq_names.CALL_TYPE:
            raise zmq_publisher_base.UnsupportedSendPattern(request.msg_type)

        push_socket = self._check_hosts_connections(
            request.target, zmq_names.socket_type_str(zmq.PULL))

        if not push_socket.connections:
            LOG.warning(_LW("Request %s was dropped because no connection")
                        % request.msg_type)
            return

        if request.msg_type in zmq_names.MULTISEND_TYPES:
            for _ in range(push_socket.connections_count()):
                self._send_request(push_socket, request)
        else:
            self._send_request(push_socket, request)

    def _send_request(self, socket, request):

        super(PushPublisher, self)._send_request(socket, request)

        LOG.debug("Publishing message %(message)s to a target %(target)s"
                  % {"message": request.message,
                     "target": request.target})
